function img = flow_viz(uv,maxlen)
%FLOW_VIZ creates a hue/saturation flow visualization
%   IMG = FLOW_VIZ(UV) converts the flow field UV to an RGB image IMG where
%   hue represents the angle of the flow vectors and saturation represents
%   the length. Maximum saturation is achieved by the longest vector in the
%   field.
%
%   IMG = FLOW_VIZ(UV,MAXLEN) sets flow vectors of length >= MAXLEN to have
%   maximum saturation.
%
% UV: Flow field with horizontal component in channel 1 and vertical
%     component in channel 2.
% MAXLEN: Vector length to use for maximum saturation.
% IMG: (Optional) Output visualization image. If no output argument is
%      specified, the image will be displayed in the current figure.

% Travis Portz
% University of Wisconsin-Madison
% 2012

u = uv(:,:,1);
v = uv(:,:,2);
[h,w] = size(u);

img = ones(h,w,3);

len = u.^2 + v.^2;
if nargin == 1
    maxlen = sqrt(max(len(:)));
end

if maxlen > 0
    for y = 1:h
        for x = 1:w
            uu = u(y,x);
            vv = v(y,x);
            theta = atan2(vv,uu);
            if theta < 0
                theta = theta + 2*pi;
            end
            img(y,x,1) = theta / (2*pi);
            img(y,x,2) = min(1,sqrt(uu^2 + vv^2) / maxlen);
        end
    end
else
    img(:,:,2) = zeros(h,w);
end

img = hsv2rgb(img);
if nargout == 0
    imshow(img);
end
