/*
 * arithmetic.hpp
 *
 *  Created on: Oct 21, 2011
 *      Author: Travis Portz
 */

#ifndef ARITHMETIC_HPP
#define ARITHMETIC_HPP

#include "image.hpp"

template <typename P>
void image_add(const image<P>& src, P p, image<P>& dst)
{
	dst.allocate(src);
	int n = src.width() * src.height();
	const P* src_data = src.data();
	P* dst_data = dst.data();
	for (int i = 0; i < n; i++)
		dst_data[i] = src_data[i] + p;
}

template <typename P>
void image_add(const image<P>& a, const image<P>& b, image<P>& dst)
{
	assert(dimensions_match(a, b));
	dst.allocate(a);
	int n = a.width() * a.height();
	const P* a_data = a.data();
	const P* b_data = b.data();
	P* dst_data = dst.data();
	for (int i = 0; i < n; i++)
		dst_data[i] = a_data[i] + b_data[i];
}

// dst = a + b * ratio
template <typename P, typename Scalar>
void image_add(const image<P>& a, const image<P>& b, Scalar ratio, image<P>& dst)
{
	assert(dimensions_match(a, b));
	dst.allocate(a);
	int n = a.width() * a.height();
	const P* a_data = a.data();
	const P* b_data = b.data();
	P* dst_data = dst.data();
	for (int i = 0; i < n; i++)
		dst_data[i] = a_data[i] + b_data[i] * ratio;
}

template <typename P>
void image_subtract(const image<P>& a, const image<P>& b, image<P>& dst)
{
	assert(dimensions_match(a, b));
	dst.allocate(a);
	int n = a.width() * a.height();
	const P* a_data = a.data();
	const P* b_data = b.data();
	P* dst_data = dst.data();
	for (int i = 0; i < n; i++)
		dst_data[i] = a_data[i] - b_data[i];
}

// P2 should be either P1 or a scalar
template <typename P1, typename P2>
void image_multiply(const image<P1>& a, const image<P2>& b, image<P1>& dst)
{
	assert(dimensions_match(a, b));
	dst.allocate(a);
	int n = a.width() * a.height();
	const P1* a_data = a.data();
	const P2* b_data = b.data();
	P1* dst_data = dst.data();
	for (int i = 0; i < n; i++)
		dst_data[i] = a_data[i] * b_data[i];
}

// dst = a + b * c
// P2 should be either P1 or a scalar
template <typename P1, typename P2>
void image_multiply_add(const image<P1>& a, const image<P1>& b, const image<P2>& c, image<P1>& dst)
{
	assert(dimensions_match(a, b) && dimensions_match(b, c));
	dst.allocate(a);
	int n = a.width() * a.height();
	const P1* a_data = a.data();
	const P1* b_data = b.data();
	const P2* c_data = c.data();
	P1* dst_data = dst.data();
	for (int i = 0; i < n; i++)
	{
		P1 p = b_data[i];
		p *= c_data[i];
		p += a_data[i];
		dst_data[i] = p;
	}
}

template <typename P>
void image_square(const image<P>& src, image<P>& dst)
{
	dst.allocate(src);
	int n = src.width() * src.height();
	const P* src_data = src.data();
	P* dst_data = dst.data();
	for (int i = 0; i < n; i++)
		dst_data[i] = src_data[i] * src_data[i];
}

// dst = a * b * c
template <typename P>
void image_multiply(const image<P>& a, const image<P>& b, const image<P>& c, image<P>& dst)
{
	assert(dimensions_match(a, b) && dimensions_match(b, c));
	dst.allocate(a);
	int n = a.width() * a.height();
	const P *a_data = a.data(), *b_data = b.data(), *c_data = c.data();
	P* dst_data = dst.data();
	for (int i = 0; i < n; i++)
	{
		P p = a_data[i];
		p *= b_data[i];
		p *= c_data[i];
		dst_data[i] = p;
	}
}

// acc = acc + b * c
template <typename P>
void image_multiply_accumulate(const image<P>& b, const image<P>& c, image<P>& acc)
{
	assert(dimensions_match(acc, b) && dimensions_match(b, c));
	int n = acc.width() * acc.height();
	P* acc_data = acc.data();
	const P *b_data = b.data(), *c_data = c.data();
	for (int i = 0; i < n; i++)
	{
		acc_data[i] += b_data[i] * c_data[i];
	}
}

template <typename Scalar>
Scalar image_norm_squared(const image<Scalar>& src)
{
	float result = 0.0f;
	int n = src.width() * src.height();
	const Scalar* src_data = src.data();
	for (int i = 0; i < n; i++)
		result += src_data[i] * src_data[i];
	return result;
}

template <typename Scalar>
Scalar image_inner_product(const image<Scalar>& a, const image<Scalar>& b)
{
	assert(dimensions_match(a, b));
	float result = 0.0f;
	int n = a.width() * a.height();
	const Scalar *a_data = a.data(), *b_data = b.data();
	for (int i = 0; i < n; i++)
		result += a_data[i] * b_data[i];
	return result;
}

#endif /* ARITHMETIC_HPP */
