/*
 * blurflow.cpp
 *
 *  Created on: Oct 21, 2011
 *      Author: Travis Portz
 */

#include "blurflow.hpp"
#include "blurgrid.hpp"
#include "filter.hpp"
#include "arithmetic.hpp"
#include "utility.hpp"
#include <sstream>

template <typename T>
T read_field(const mxArray* params, const char* name, T default_value)
{
	int field_num = mxGetFieldNumber(params, name);
	if (field_num >= 0)
	{
		mxArray* field = mxGetFieldByNumber(params, 0, field_num);
		return static_cast<T>(*static_cast<double*>(mxGetData(field)));
	}
	return default_value;
}

flow_params::flow_params(const mxArray* params)
{
	scale_factor = read_field(params, "scale_factor", 0.75f);
	alpha = read_field(params, "alpha", 0.01f);
	gamma = read_field(params, "gamma", 1.0f);
	epsilon_data = read_field(params, "epsilon_data", 0.001f);
	epsilon_spatial = read_field(params, "epsilon_spatial", 0.001f);
	num_pyramid_levels = read_field(params, "num_pyramid_levels", 4);
	max_outer_iterations = read_field(params, "max_outer_iterations", 16);
	max_inner_iterations = read_field(params, "max_inner_iterations", 1);
	max_cg_iterations = read_field(params, "max_cg_iterations", 50);
	cg_threshold = read_field(params, "cg_threshold", 1e-8f);
	max_images = read_field(params, "max_images", 240);
	color_weight = read_field(params, "color_weight", 0.5f);
	int field_num = mxGetFieldNumber(params, "range");
	if (field_num >= 0)
	{
		mxArray* field = mxGetFieldByNumber(params, 0, field_num);
		double* data = static_cast<double*>(mxGetData(field));
		range.u_min = static_cast<float>(data[0]);
		range.u_max = static_cast<float>(data[1]);
		range.v_min = static_cast<float>(data[2]);
		range.v_max = static_cast<float>(data[3]);
	}
	else
		range.u_min = range.u_max = range.v_min = range.v_max = 0.0f;
}

void generate_feature_image(const image3f& src, feature_image& dst, float gamma, float color_weight)
{
	dst.allocate(src);

	image1f gray;
	desaturate(src, gray);

	image1f dx, dy;
	filter_x(gray, gradient_kernel, dx);
	filter_y(gray, gradient_kernel, dy);

	float sqrt_gamma = std::sqrt(gamma);
	float sqrt_color = std::sqrt(color_weight);

	const float *gray_data = gray.data(), *dx_data = dx.data(), *dy_data = dy.data();
	const pixel3f* src_data = src.data();
	feature_pixel* dst_data = dst.data();

	int n = dst.width() * dst.height();
	for (int i = 0; i < n; i++)
	{
		pixel3f s = src_data[i];
		feature_pixel& p = dst_data[i];
		p[0] = gray_data[i];
		p[1] = sqrt_gamma * dx_data[i]; // sqrt(gamma) * dy
		p[2] = sqrt_gamma * dy_data[i]; // sqrt(gamma) * dx
		// (p[1]^2 + p[2]^2) = gamma * (dx^2 + dy^2)
		p[3] = sqrt_color * (s[1] - s[0]); // green - red
		p[4] = sqrt_color * (s[1] - s[2]); // green - blue
		dst_data[i] = p;
	}
}

void warp(const feature_image& src, const image1f& u, const image1f& v, feature_image& dst)
{
	int width = src.width(), height = src.height();
	dst.allocate(width, height);
	float xu_max = width - 1.0f, yv_max = height - 1.0f;
	for (int y = 0; y < height; y++)
	{
		const float *u_row = u.row(y), *v_row = v.row(y);
		feature_pixel *dst_row = dst.row(y);
		for (int x = 0; x < width; x++)
		{
			float u = u_row[x];
			float v = v_row[x];
			float xu = clamp(x + u, 0.0f, xu_max), yv = clamp(y + v, 0.0f, yv_max);
			dst_row[x] = src.interpolate(xu, yv);
		}
	}
}

void compute_flow_baseline(image3f& from, image3f& to, flow_params params, flow32f_field& field)
{
	image3f* from_pyramid = new image3f[params.num_pyramid_levels];
	from_pyramid[0].deep_copy(from);
	build_pyramid(from_pyramid, params.num_pyramid_levels, params.scale_factor);

	image3f* to_pyramid = new image3f[params.num_pyramid_levels];
	to_pyramid[0].deep_copy(to);
	build_pyramid(to_pyramid, params.num_pyramid_levels, params.scale_factor);

	float eps_d_squared = params.epsilon_data * params.epsilon_data;
	float eps_s_squared = params.epsilon_spatial * params.epsilon_spatial;

	image1f u, v;

	int max_level = params.num_pyramid_levels - 1;
	for (int level = max_level; level >= 0; level--)
	{
		log_message("level %d", level);

		feature_image im1, im2;
		generate_feature_image(from_pyramid[level], im1, params.gamma, params.color_weight);
		generate_feature_image(to_pyramid[level], im2, params.gamma, params.color_weight);

		if (level < max_level)
		{
			// Upsample and scale flow.
			image1f u_new;
			resize(u, u_new, im1.width(), im1.height());
			float x_scale = u_new.width() / static_cast<float>(u.width());
			u_new.multiply_with(x_scale);
			u.swap_with(u_new);

			image1f v_new;
			resize(v, v_new, im1.width(), im1.height());
			float y_scale = v_new.height() / static_cast<float>(v.height());
			v_new.multiply_with(y_scale);
			v.swap_with(v_new);
		}
		else
		{
			// Initialize flow.
			u.allocate(im1);
			u.zero();
			v.allocate(im1);
			v.zero();
		}

		feature_image dxI2, dyI2;
		filter_x(im2, gradient_kernel, dxI2);
		filter_y(im2, gradient_kernel, dyI2);

		feature_image I2_warped;
		feature_image Iz, Ix, Iy;

		int w = u.width(), h = u.height();
		int n = w * h;

		image1f du(w, h), dv(w, h);

		image1f uu, vv; // uu = u + du, vv = v + dv
		image1f ux, uy, vx, vy; // ux = dx(u + du), uy = dy(u + du), ...

		feature_image dcost_d(w, h);
		image1f dcost_s(w, h);

		image1f Auu, Avv, Auv, bu, bv;
		image1f Lu, Lv;

		image1f ru, rv, pu, pv, qu, qv;

		feature_image temp;
		image1f collapsed;

		for (int outer = 0; outer < params.max_outer_iterations; outer++)
		{
			log_message(".");

			warp(im2, u, v, I2_warped);
			image_subtract(I2_warped, im1, Iz); // Iz = I2(x+u,y+v) - I1(x,y)

			warp(dxI2, u, v, Ix); // Ix = dxI2(x+u,y+v)
			warp(dyI2, u, v, Iy); // Iy = dyI2(x+u,y+v)

			du.zero(); dv.zero();

			for (int inner = 0; inner < params.max_inner_iterations; inner++)
			{
				image_add(u, du, uu);
				image_add(v, dv, vv);
				simple_dx(uu, ux); simple_dy(uu, uy);
				simple_dx(vv, vx); simple_dy(vv, vy);

				// Compute data cost function derivative
				feature_pixel *dcost_d_data = dcost_d.data();
				const feature_pixel *Iz_data = Iz.data(), *Ix_data = Ix.data(), *Iy_data = Iy.data();
				const float *du_data = du.data(), *dv_data = dv.data();
				for (int i = 0; i < n; i++)
				{
					feature_pixel f(Iz_data[i]);
					f += Ix_data[i] * du_data[i];
					f += Iy_data[i] * dv_data[i];
					feature_pixel& p = dcost_d_data[i];
					for (int c = 0; c < D; c++)
						p[c] = 0.5f / std::sqrt(f[c] * f[c] + eps_d_squared);
				}

				// Compute spatial cost function derivative
				float *dcost_s_data = dcost_s.data();
				const float *ux_data = ux.data(), *uy_data = uy.data(), *vx_data = vx.data(), *vy_data = vy.data();
				for (int i = 0; i < n; i++)
				{
					float d = ux_data[i], g2 = d * d;
					d = uy_data[i]; g2 += d * d;
					d = vx_data[i]; g2 += d * d;
					d = vy_data[i]; g2 += d * d;
					dcost_s_data[i] = 0.5f / std::sqrt(g2 + eps_s_squared);
				}

				// Compute A matrix

				image_multiply(Ix, Ix, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, Auu);
				Auu.add(params.alpha * 0.01f);

				image_multiply(Iy, Iy, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, Avv);
				Avv.add(params.alpha * 0.01f);

				image_multiply(Ix, Iy, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, Auv);

				// Compute b vector

				image_multiply(Ix, Iz, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, bu);

				image_multiply(Iy, Iz, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, bv);

				weighted_laplacian(u, dcost_s, Lu);
				weighted_laplacian(v, dcost_s, Lv);

				const float *Lu_data = Lu.data(), *Lv_data = Lv.data();
				float *bu_data = bu.data(), *bv_data = bv.data();
				for (int i = 0; i < n; i++)
				{
					bu_data[i] = -bu_data[i] - params.alpha * Lu_data[i];
					bv_data[i] = -bv_data[i] - params.alpha * Lv_data[i];
				}

				// Begin conjugate gradient method

				du.zero();
				dv.zero();
				ru.deep_copy(bu);
				rv.deep_copy(bv);

				float R = image_norm_squared(ru) + image_norm_squared(rv);
				float R_prev = -1.0f;
				
				for (int cg = 0; cg < params.max_cg_iterations && R > params.cg_threshold; cg++)
				{
					if (cg == 0)
					{
						pu.deep_copy(ru);
						pv.deep_copy(rv);
					}
					else
					{
						assert(R_prev > 0.0f);
						// p = r + (R / R_prev) * p
						float ratio = R / R_prev;
						image_add(ru, pu, ratio, pu);
						image_add(rv, pv, ratio, pv);
					}

					// q = A * p
					image_multiply(Auu, pu, qu);
					image_multiply_accumulate(Auv, pv, qu);
					weighted_laplacian(pu, dcost_s, Lu);
					qu.add(Lu, params.alpha);

					image_multiply(Auv, pu, qv);
					image_multiply_accumulate(Avv, pv, qv);
					weighted_laplacian(pv, dcost_s, Lv);
					qv.add(Lv, params.alpha);

					// lambda = R / (p' * q)
					float lambda = R / (image_inner_product(pu, qu) + image_inner_product(pv, qv));

					// x = x + lambda * p
					du.add(pu, lambda);
					dv.add(pv, lambda);

					// r = r - lambda * q
					ru.add(qu, -lambda);
					rv.add(qv, -lambda);

					R_prev = R;
					R = image_norm_squared(ru) + image_norm_squared(rv);
				} // Conjugate gradient
			} // Inner iteration
			u.add(du);
			v.add(dv);
		} // Outer iteration
		log_message("\n");
	} // Pyramid level

	log_message("writing flow\n");
	flow_from_images(u, v, field);

	delete [] from_pyramid;
	delete [] to_pyramid;
} // compute_flow_baseline

void compute_flow_blurred(image3f& from, image3f& to, float tau1, float tau2, flow_params params,
		const flow32f_field& flow10, const flow32f_field& flow12, const flow32f_field& flow23, flow32f_field& field)
{
	image3f* from_pyramid = new image3f[params.num_pyramid_levels];
	from_pyramid[0].deep_copy(from);
	build_pyramid(from_pyramid, params.num_pyramid_levels, params.scale_factor);

	image3f* to_pyramid = new image3f[params.num_pyramid_levels];
	to_pyramid[0].deep_copy(to);
	build_pyramid(to_pyramid, params.num_pyramid_levels, params.scale_factor);

	float eps_d_squared = params.epsilon_data * params.epsilon_data;
	float eps_s_squared = params.epsilon_spatial * params.epsilon_spatial;

	image1f u, v;

	int max_level = params.num_pyramid_levels - 1;
	for (int level = max_level; level >= 0; level--)
	{
		log_message("level %d\n", level);

		feature_image im1, im2;
		generate_feature_image(from_pyramid[level], im1, params.gamma, params.color_weight);
		generate_feature_image(to_pyramid[level], im2, params.gamma, params.color_weight);

		float scale = std::pow(params.scale_factor, level);
		blur_range range = params.range;
		range.u_min *= scale; range.u_max *= scale; range.v_min *= scale; range.v_max *= scale;

		flow32f_field flow23_level, flow10_level;
		flow_resize(flow10, flow10_level, im1.width(), im1.height());
		flow_resize(flow23, flow23_level, im1.width(), im1.height());

		log_message("building image 1 grid 1... ");
		blurred_grid_11<feature_image> grid11(im1, tau2, range, params.max_images);
		log_message("image 1 grid 2... ");
		blurred_grid_12<feature_image> grid12(im1, tau2, flow23_level, params.max_images);
		log_message("done\n");

		log_message("building image 2 grid 1... ");
		blurred_grid_21<feature_image> grid21(im2, tau1, flow10_level, params.max_images);
		log_message("image 2 grid 2... ");
		blurred_grid_22<feature_image> grid22(im2, tau1, range, params.max_images);
		log_message("optimizing flow");

		if (level < max_level)
		{
			// Upsample and scale flow.
			image1f u_new;
			resize(u, u_new, im1.width(), im1.height());
			float x_scale = u_new.width() / static_cast<float>(u.width());
			u_new.multiply_with(x_scale);
			u.swap_with(u_new);

			image1f v_new;
			resize(v, v_new, im1.width(), im1.height());
			float y_scale = v_new.height() / static_cast<float>(v.height());
			v_new.multiply_with(y_scale);
			v.swap_with(v_new);
		}
		else
		{
			// Initialize flow.
			flow32f_field initial_flow;
			flow_resize(flow12, initial_flow, im1.width(), im1.height());
			flow_to_images(initial_flow, u, v);
		}

		feature_image B_half, du_half, dv_half, dx_half, dy_half;
		feature_image B1, duB1, dvB1;
		feature_image B2, duB2, dvB2, dxB2, dyB2;
		feature_image Bz, Buz, Bvz;

		int w = u.width(), h = u.height();
		int n = w * h;

		image1f du(w, h), dv(w, h);

		image1f uu, vv; // uu = u + du, vv = v + dv
		image1f ux, uy, vx, vy; // ux = dx(u + du), uy = dy(u + du), ...

		feature_image dcost_d(w, h);
		image1f dcost_s(w, h);

		image1f Auu, Avv, Auv, bu, bv;
		image1f Lu, Lv;

		image1f ru, rv, pu, pv, qu, qv;

		feature_image temp;
		image1f collapsed;

		for (int outer = 0; outer < params.max_outer_iterations; outer++)
		{
			log_message(".");

			// Bz = B2 - B1 = (B21 + B22)/2 - (B11 + B12)/2
			// Buz = duB2 + dxB2 - duB1 = (dxB21 + duB22 + dxB22)/2 - (duB11 + duB22)/2
			// Bvz = dvB2 + dyB2 - dvB1 = (dyB21 + dvB22 + dyB22)/2 - (dvB11 + dvB22)/2

			grid11.blur(u, v, B1, duB1, dvB1);
			grid12.blur(u, v, B_half, du_half, dv_half);
			B1.average_with(B_half);
			duB1.average_with(du_half); dvB1.average_with(dv_half);

			grid21.blur(u, v, B2, dxB2, dyB2);
			grid22.blur(u, v, B_half, duB2, dvB2, dx_half, dy_half);
			B2.average_with(B_half);
			duB2.multiply_with(0.5f); dvB2.multiply_with(0.5f); // average with duB21 = 0, dvB21 = 0
			dxB2.average_with(dx_half); dyB2.average_with(dy_half);

			image_subtract(B2, B1, Bz);                     // Bz   = B2 - B1
			image_add(duB2, dxB2, Buz); Buz.subtract(duB1); // Buz  = duB2 + dxB2 - duB1
			image_add(dvB2, dyB2, Bvz); Bvz.subtract(dvB1); // Bvz  = dvB2 + dyB2 - dvB1

			du.zero(); dv.zero();

			for (int inner = 0; inner < params.max_inner_iterations; inner++)
			{
				image_add(u, du, uu);
				image_add(v, dv, vv);
				simple_dx(uu, ux); simple_dy(uu, uy);
				simple_dx(vv, vx); simple_dy(vv, vy);

				// Compute data cost function derivative
				feature_pixel *dcost_d_data = dcost_d.data();
				const feature_pixel *Bz_data = Bz.data(), *Buz_data = Buz.data(), *Bvz_data = Bvz.data();
				const float *du_data = du.data(), *dv_data = dv.data();
				for (int i = 0; i < n; i++)
				{
					feature_pixel f(Bz_data[i]);
					f += Buz_data[i] * du_data[i];
					f += Bvz_data[i] * dv_data[i];
					feature_pixel& p = dcost_d_data[i];
					for (int c = 0; c < D; c++)
						p[c] = 0.5f / std::sqrt(f[c] * f[c] + eps_d_squared);
				}

				// Compute spatial cost function derivative
				float *dcost_s_data = dcost_s.data();
				const float *ux_data = ux.data(), *uy_data = uy.data(), *vx_data = vx.data(), *vy_data = vy.data();
				for (int i = 0; i < n; i++)
				{
					float d = ux_data[i], g2 = d * d;
					d = uy_data[i]; g2 += d * d;
					d = vx_data[i]; g2 += d * d;
					d = vy_data[i]; g2 += d * d;
					dcost_s_data[i] = 0.5f / std::sqrt(g2 + eps_s_squared);
				}

				// Compute A matrix

				image_multiply(Buz, Buz, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, Auu);
				Auu.add(params.alpha * 0.01f);

				image_multiply(Bvz, Bvz, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, Avv);
				Avv.add(params.alpha * 0.01f);

				image_multiply(Buz, Bvz, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, Auv);

				// Compute b vector

				image_multiply(Buz, Bz, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, bu);

				image_multiply(Bvz, Bz, dcost_d, temp);
				collapse(temp, collapsed);
				filter_xy(collapsed, smoothing_kernel, bv);

				weighted_laplacian(u, dcost_s, Lu);
				weighted_laplacian(v, dcost_s, Lv);

				const float *Lu_data = Lu.data(), *Lv_data = Lv.data();
				float *bu_data = bu.data(), *bv_data = bv.data();
				for (int i = 0; i < n; i++)
				{
					bu_data[i] = -bu_data[i] - params.alpha * Lu_data[i];
					bv_data[i] = -bv_data[i] - params.alpha * Lv_data[i];
				}

				// Begin conjugate gradient method

				du.zero();
				dv.zero();
				ru.deep_copy(bu);
				rv.deep_copy(bv);

				float R = image_norm_squared(ru) + image_norm_squared(rv);
				float R_prev = -1.0f;
				
				for (int cg = 0; cg < params.max_cg_iterations && R > params.cg_threshold; cg++)
				{
					if (cg == 0)
					{
						pu.deep_copy(ru);
						pv.deep_copy(rv);
					}
					else
					{
						assert(R_prev > 0.0f);
						// p = r + (R / R_prev) * p
						float ratio = R / R_prev;
						image_add(ru, pu, ratio, pu);
						image_add(rv, pv, ratio, pv);
					}

					// q = A * p
					image_multiply(Auu, pu, qu);
					image_multiply_accumulate(Auv, pv, qu);
					weighted_laplacian(pu, dcost_s, Lu);
					qu.add(Lu, params.alpha);

					image_multiply(Auv, pu, qv);
					image_multiply_accumulate(Avv, pv, qv);
					weighted_laplacian(pv, dcost_s, Lv);
					qv.add(Lv, params.alpha);

					// lambda = R / (p' * q)
					float lambda = R / (image_inner_product(pu, qu) + image_inner_product(pv, qv));

					// x = x + lambda * p
					du.add(pu, lambda);
					dv.add(pv, lambda);

					// r = r - lambda * q
					ru.add(qu, -lambda);
					rv.add(qv, -lambda);

					R_prev = R;
					R = image_norm_squared(ru) + image_norm_squared(rv);
				} // Conjugate gradient
			} // Inner iteration
			u.add(du);
			v.add(dv);
		} // Outer iteration
		log_message("\n");
	} // Pyramid level

	log_message("writing flow\n");
	flow_from_images(u, v, field);

	delete [] from_pyramid;
	delete [] to_pyramid;
} // compute_flow_blurred

