/*
 * blurgrid.hpp
 *
 *  Created on: Nov 3, 2011
 *      Author: Travis Portz
 */

#ifndef BLURGRID_HPP
#define BLURGRID_HPP

#include "blur.hpp"

// For two-sided motion estimate.
template <class image_type>
class blurred_grid
{
public:
	typedef typename image_type::pixel_type pixel_type;

	// src: the image being blurred, it is not const because we will make shallow copies for any unblurred points in the grid
	// tau: the duty cycle of the other image whose blur we are trying to apply to src
	blurred_grid(image_type& src, float tau, blur_range range, int max_images, bool gradients);
	~blurred_grid() { delete [] images; delete [] dx_images; delete [] dy_images; }

	void operator()(int i, int j);

protected:
	image_type& src;
	std::vector<float> u_points;
	std::vector<float> v_points;
	image_type* images;
	image_type* dx_images;
	image_type* dy_images;
	float tau;
	blur_range_mode mode;
	int zero_offset;
	bool gradients;

	void locate_cell(float u, float v, blurred_image_cell& cell);
	void compute_grid_points(const blur_range& range, int max_images);

	inline void adjust_coord(float& u, float& v)
	{
		switch (mode)
		{
		case blur_positive_v:
			if (v < 0) { u = -u; v = -v; }
			break;
		case blur_positive_u:
			if (u < 0) { u = -u; v = -v; }
			break;
		case blur_full_range:
			break;
		}
	}
};

// B11(z) = (I1 * k[-w12(z)tau2/2])(z)
template <class image_type>
class blurred_grid_11 : public blurred_grid<image_type>
{
public:
	typedef typename image_type::pixel_type pixel_type;
	typedef blurred_grid<image_type> base;

	blurred_grid_11(image_type& src, float tau, blur_range range12, int max_images);
	void blur(const image1f& u, const image1f& v, image_type& B, image_type& duB, image_type& dvB);
};

// B12(z) = (I1 * k[w23(z + w12(z))tau2/2])(z)
template <class image_type>
class blurred_grid_12 : public blurred_grid<image_type>
{
public:
	typedef typename image_type::pixel_type pixel_type;
	typedef blurred_grid<image_type> base;

	blurred_grid_12(image_type& src, float tau, const flow32f_field& flow23, int max_images);
	void blur(const image1f& u, const image1f& v, image_type& B, image_type& duB, image_type& dvB);

private:
	typedef pixel<float, 6> flow_pixel; // (u, v, du/dx, du/dy, dv/dx, dv/dy)
	typedef image<flow_pixel> flow_image;
	flow_image flow23;
};

// B21(z) = (I2 * k[w10(z)tau1/2])(z + w12(z))
template <class image_type>
class blurred_grid_21 : public blurred_grid<image_type>
{
public:
	typedef typename image_type::pixel_type pixel_type;
	typedef blurred_grid<image_type> base;

	blurred_grid_21(image_type& src, float tau, const flow32f_field& flow10, int max_images);
	void blur(const image1f& u, const image1f& v, image_type& B, image_type& dxB, image_type& dyB);

private:
	const flow32f_field& flow10;
};

// B22(z) = (I2 * k[w12(z)tau1/2])(z + w12(z))
template <class image_type>
class blurred_grid_22 : public blurred_grid<image_type>
{
public:
	typedef typename image_type::pixel_type pixel_type;
	typedef blurred_grid<image_type> base;

	blurred_grid_22(image_type& src, float tau, blur_range range, int max_images);
	void blur(const image1f& u, const image1f& v, image_type& B, image_type& duB, image_type& dvB, image_type& dxB, image_type& dyB);
};

#include "blurgrid_detail.hpp"

#endif /* BLURGRID_HPP */
