/*
 * blurgrid_detail.hpp
 *
 *  Created on: Nov 3, 2011
 *      Author: Travis Portz
 */

#ifndef BLURGRID_DETAIL_HPP
#define BLURGRID_DETAIL_HPP

#include "blurgrid.hpp"
#include "poly.hpp"
#include "filter.hpp"
#include <omp.h>
#include <sstream>

template <class image_type>
blurred_grid<image_type>::blurred_grid(image_type& src, float tau_other, blur_range range, int max_images, bool gradients)
	: src(src), tau(tau_other * 0.5f), mode(range.mode), zero_offset(-1), gradients(gradients)
{
	range.u_min *= tau; range.u_max *= tau; range.v_min *= tau; range.v_max *= tau;
	compute_grid_points(range, max_images);

	int w = u_points.size(), h = v_points.size();
	int n = w * h;
	images = new image_type[n];

	if (gradients)
	{
		dx_images = new image_type[n];
		dy_images = new image_type[n];
	}
	else
	{
		dx_images = nullptr;
		dy_images = nullptr;
	}

	#pragma omp parallel for schedule(static, 1)
	for (int i = 0; i < n; i++)
	{
		int x = i % w;
		int y = i / w;
		(*this)(x, y);
	}
}

template <class image_type>
void blurred_grid<image_type>::operator()(int i, int j)
{
	int w = u_points.size();

	float u = u_points[i];
	float v = v_points[j];
	int offset = j * w + i;
	float length = std::sqrt(u * u + v * v);
	if (length > 0.5f)
	{
		kernel2d kern;
		motion_blur_kernel_oc(u, v, kern);
		filter_xy_oc(src, kern, u, v, images[offset]);
		if (gradients)
		{
			filter_x(images[offset], gradient_kernel, dx_images[offset]);
			filter_y(images[offset], gradient_kernel, dy_images[offset]);
		}
	}
	else
	{
		// Blur kernel is a single point, just use the original image.
		images[offset].shallow_copy(src);
		if (gradients)
		{
			if (zero_offset < 0)
			{
				filter_x(src, gradient_kernel, dx_images[offset]);
				filter_y(src, gradient_kernel, dy_images[offset]);
				zero_offset = offset;
			}
			else
			{
				dx_images[offset].shallow_copy(dx_images[zero_offset]);
				dy_images[offset].shallow_copy(dy_images[zero_offset]);
			}
		}
	}
}

template <class image_type>
void blurred_grid<image_type>::locate_cell(float u, float v, blurred_image_cell& cell)
{
	typedef std::vector<float>::iterator point_iterator;
	point_iterator u_iter = std::lower_bound(u_points.begin(), u_points.end(), u);
	if (u_iter >= u_points.end() - 1)
		u_iter = u_points.end() - 2;
	cell.i = u_iter - u_points.begin();
	cell.u1 = *u_iter; cell.u2 = u_iter[1];
	point_iterator v_iter = std::lower_bound(v_points.begin(), v_points.end(), v);
	if (v_iter >= v_points.end() - 1)
		v_iter = v_points.end() - 2;
	cell.j = v_iter - v_points.begin();
	cell.v1 = *v_iter; cell.v2 = v_iter[1];
	cell.offset = cell.j * u_points.size() + cell.i;
	cell.coeff = 1.0f / ((cell.u2 - cell.u1) * (cell.v2 - cell.v1));
}


template <class image_type>
void blurred_grid<image_type>::compute_grid_points(const blur_range& range, int max_images)
{
	float u_min = range.u_min, u_max = range.u_max, v_min = range.v_min, v_max = range.v_max;
	float w = u_max - u_min, h = v_max - v_min;
	float aspect = w / h;
	int n = std::min(max_images, std::max(1, static_cast<int>(std::sqrt(max_images * aspect))));
	int m = max_images / n;

	compute_axis_points(u_min, u_max, n, u_points);
	compute_axis_points(v_min, v_max, m, v_points);
}

// Flow is opposite for this version, so negate the blur range.
template <class image_type>
blurred_grid_11<image_type>::blurred_grid_11(image_type& src, float tau, blur_range range12, int max_images)
	: blurred_grid<image_type>(src, tau, range12.negate(), max_images, false) { }

template <class image_type>
void blurred_grid_11<image_type>::blur(const image1f& u, const image1f& v, image_type& B, image_type& duB, image_type& dvB)
{
	assert(dimensions_match(u, v));
	int width = u.width(), height = u.height();
	B.allocate(width, height); duB.allocate(width, height); dvB.allocate(width, height);

	for (int y = 0; y < height; y++)
	{
		const float *u_row = u.row(y), *v_row = v.row(y);
		pixel_type *B_row = B.row(y), *duB_row = duB.row(y), *dvB_row = dvB.row(y);
		blurred_image_cell cell;
		pixel_type *B11 = nullptr, *B21= nullptr, *B12 = nullptr, *B22 = nullptr;
		int pixel_offset = y * width;
		for (int x = 0; x < width; x++)
		{
			// Flow is opposite for this version.
			float u = -u_row[x] * base::tau;
			float v = -v_row[x] * base::tau;
			base::adjust_coord(u, v);

			if (cell.is_outside(u, v))
			{
				base::locate_cell(u, v, cell);
				B11 = base::images[cell.offset].data();
				B21 = base::images[cell.offset + 1].data();
				B12 = base::images[cell.offset + base::u_points.size()].data();
				B22 = base::images[cell.offset + base::u_points.size() + 1].data();
			}

			u = clamp(u, cell.u1, cell.u2);
			v = clamp(v, cell.v1, cell.v2);

			float s1 = cell.u2 - u, s2 = u - cell.u1;
			float t1 = cell.v2 - v, t2 = v - cell.v1;

			pixel_type b11 = B11[pixel_offset], b21 = B21[pixel_offset], b12 = B12[pixel_offset], b22 = B22[pixel_offset];

			pixel_type b = b11 * (s1 * t1);
			b += b21 * (s2 * t1);
			b += b12 * (s1 * t2);
			b += b22 * (s2 * t2);
			B_row[x] = b * cell.coeff;

			pixel_type du = (b21 - b11) * t1;
			du += (b22 - b12) * t2;
			duB_row[x] = -du * (cell.coeff * base::tau); // Negate derivative since flow is opposite.

			pixel_type dv = (b12 - b11) * s1;
			dv += (b22 - b21) * s2;
			dvB_row[x] = -dv * (cell.coeff * base::tau); // Negate derivative since flow is opposite.

			pixel_offset++;
		}
	}
}

template <class image_type>
blurred_grid_12<image_type>::blurred_grid_12(image_type& src, float tau, const flow32f_field& field23, int max_images)
	: blurred_grid<image_type>(src, tau, blur_range::from_field(field23), max_images, false), flow23(field23.width(), field23.height())
{
	image1f u, v;
	flow_to_images(field23, u, v);
	image1f dudx, dudy, dvdx, dvdy;
	filter_x(u, gradient_kernel, dudx);
	filter_y(u, gradient_kernel, dudy);
	filter_x(v, gradient_kernel, dvdx);
	filter_y(v, gradient_kernel, dvdy);

	const float *u_data = u.data(), *v_data = v.data(),
			*dudx_data = dudx.data(), *dudy_data = dudy.data(),
			*dvdx_data = dvdx.data(), *dvdy_data = dvdy.data();
	flow_pixel* dst = flow23.data();
	int n = field23.width() * field23.height();
	for (int i = 0; i < n; i++)
	{
		flow_pixel p;
		p[0] = u_data[i]; p[1] = v_data[i];
		p[2] = dudx_data[i]; p[3] = dudy_data[i];
		p[4] = dvdx_data[i]; p[5] = dvdy_data[i];
		dst[i] = p;
	}
}

template <class image_type>
void blurred_grid_12<image_type>::blur(const image1f& u, const image1f& v, image_type& B, image_type& duB, image_type& dvB)
{
	assert(dimensions_match(u, v));
	int width = u.width(), height = u.height();
	B.allocate(width, height); duB.allocate(width, height); dvB.allocate(width, height);

	float xu_max = width - 1.0f, yv_max = height - 1.0f;

	for (int y = 0; y < height; y++)
	{
		const float *u_row = u.row(y), *v_row = v.row(y);
		pixel_type *B_row = B.row(y), *duB_row = duB.row(y), *dvB_row = dvB.row(y);
		blurred_image_cell cell;
		pixel_type *B11 = nullptr, *B21= nullptr, *B12 = nullptr, *B22 = nullptr;
		int pixel_offset = y * width;
		for (int x = 0; x < width; x++)
		{
			// Blur kernel is k[w23(z + w12(z))tau2/2].
			float u = u_row[x], v = v_row[x]; // w12(z)
			float xu = clamp(x + u, 0.0f, xu_max), yv = clamp(y + v, 0.0f, yv_max); // z + w12(z)
			flow_pixel flow = flow23.interpolate(xu, yv); // w23(z + w12(z))
			u = flow[0] * base::tau; v = flow[1] * base::tau; // w23(z + w12(z))tau2/2
			base::adjust_coord(u, v);

			if (cell.is_outside(u, v))
			{
				base::locate_cell(u, v, cell);
				B11 = base::images[cell.offset].data();
				B21 = base::images[cell.offset + 1].data();
				B12 = base::images[cell.offset + base::u_points.size()].data();
				B22 = base::images[cell.offset + base::u_points.size() + 1].data();
			}

			u = clamp(u, cell.u1, cell.u2);
			v = clamp(v, cell.v1, cell.v2);

			float s1 = cell.u2 - u, s2 = u - cell.u1;
			float t1 = cell.v2 - v, t2 = v - cell.v1;

			pixel_type b11 = B11[pixel_offset], b21 = B21[pixel_offset], b12 = B12[pixel_offset], b22 = B22[pixel_offset];

			pixel_type b = b11 * (s1 * t1);
			b += b21 * (s2 * t1);
			b += b12 * (s1 * t2);
			b += b22 * (s2 * t2);
			B_row[x] = b * cell.coeff;

			// Blur grid is in w23 coordinates, we want derivatives with respect to w12.
			// w12 is used to warp w23, so we have dB/du12 = (dB/du23 * du23/dx + dB/dv23 * dv23/dx) * dx/du12
			// and dB/dv12 = (dB/du23 * dv23/dy + dB/dv23 * dv23/dy) * dy/dv12,
			// where dx/du12 = dy/dv12 = 1.

			pixel_type dB_du23 = (b21 - b11) * t1;
			dB_du23 += (b22 - b12) * t2;
			dB_du23 *= cell.coeff * base::tau;

			pixel_type dB_dv23 = (b12 - b11) * s1;
			dB_dv23 += (b22 - b21) * s2;
			dB_dv23 *= cell.coeff * base::tau;

			pixel_type dB_du12 = dB_du23 * flow[2]; // flow[2] = du23/dx
			dB_du12 += dB_dv23 * flow[4]; // flow[4] = dv23/dx
			duB_row[x] = dB_du12;

			pixel_type dB_dv12 = dB_du23 * flow[3]; // flow[3] = du23/dy
			dB_dv12 += dB_dv23 * flow[5]; // flow[5] = dv23/dy
			dvB_row[x] = dB_dv12;

			pixel_offset++;
		}
	}
}

template <class image_type>
blurred_grid_21<image_type>::blurred_grid_21(image_type& src, float tau, const flow32f_field& flow10, int max_images)
	: blurred_grid<image_type>(src, tau, blur_range::from_field(flow10), max_images, true), flow10(flow10) { }

template <class image_type>
void blurred_grid_21<image_type>::blur(const image1f& u, const image1f& v, image_type& B, image_type& dxB, image_type& dyB)
{
	assert(dimensions_match(u, v));
	int width = u.width(), height = u.height();
	B.allocate(width, height);
	dxB.allocate(width, height); dyB.allocate(width, height);

	float xu_max = width - 1.0f, yv_max = height - 1.0f;

	for (int y = 0; y < height; y++)
	{
		const float *u_row = u.row(y), *v_row = v.row(y);
		pixel_type *B_row = B.row(y);
		pixel_type *dxB_row = dxB.row(y), *dyB_row = dyB.row(y);
		blurred_image_cell cell;
		image_type *B11 = nullptr, *B21= nullptr, *B12 = nullptr, *B22 = nullptr;
		image_type *dxB11 = nullptr, *dxB21= nullptr, *dxB12 = nullptr, *dxB22 = nullptr;
		image_type *dyB11 = nullptr, *dyB21= nullptr, *dyB12 = nullptr, *dyB22 = nullptr;
		for (int x = 0; x < width; x++)
		{
			float u = u_row[x], v = v_row[x]; // w12(z)
			float xu = clamp(x + u, 0.0f, xu_max), yv = clamp(y + v, 0.0f, yv_max); // z + w12(z), used to warp image

			// Blur kernel is k[w10(z)tau1/2].
			flow32f flow = flow10.at(x, y); // w10(z)
			u = flow.x * base::tau; v = flow.y * base::tau; // w10(z)tau1/2
			base::adjust_coord(u, v);

			if (cell.is_outside(u, v))
			{
				base::locate_cell(u, v, cell);
				B11 = base::images + cell.offset;
				B21 = base::images + cell.offset + 1;
				B12 = base::images + cell.offset + base::u_points.size();
				B22 = base::images + cell.offset + base::u_points.size() + 1;
				dxB11 = base::dx_images + cell.offset;
				dxB21 = base::dx_images + cell.offset + 1;
				dxB12 = base::dx_images + cell.offset + base::u_points.size();
				dxB22 = base::dx_images + cell.offset + base::u_points.size() + 1;
				dyB11 = base::dy_images + cell.offset;
				dyB21 = base::dy_images + cell.offset + 1;
				dyB12 = base::dy_images + cell.offset + base::u_points.size();
				dyB22 = base::dy_images + cell.offset + base::u_points.size() + 1;
			}

			u = clamp(u, cell.u1, cell.u2);
			v = clamp(v, cell.v1, cell.v2);

			float s1 = cell.u2 - u, s2 = u - cell.u1;
			float t1 = cell.v2 - v, t2 = v - cell.v1;

			pixel_type b11 = B11->interpolate(xu, yv), b21 = B21->interpolate(xu, yv);
			pixel_type b12 = B12->interpolate(xu, yv), b22 = B22->interpolate(xu, yv);

			pixel_type b = b11 * (s1 * t1);
			b += b21 * (s2 * t1);
			b += b12 * (s1 * t2);
			b += b22 * (s2 * t2);
			B_row[x] = b * cell.coeff;

			pixel_type dx11 = dxB11->interpolate(xu, yv), dx21 = dxB21->interpolate(xu, yv);
			pixel_type dx12 = dxB12->interpolate(xu, yv), dx22 = dxB22->interpolate(xu, yv);
			pixel_type dx = dx11 * (s1 * t1);
			dx += dx21 * (s2 * t1);
			dx += dx12 * (s1 * t2);
			dx += dx22 * (s2 * t2);
			dxB_row[x] = dx * cell.coeff;

			pixel_type dy11 = dyB11->interpolate(xu, yv), dy21 = dyB21->interpolate(xu, yv);
			pixel_type dy12 = dyB12->interpolate(xu, yv), dy22 = dyB22->interpolate(xu, yv);
			pixel_type dy = dy11 * (s1 * t1);
			dy += dy21 * (s2 * t1);
			dy += dy12 * (s1 * t2);
			dy += dy22 * (s2 * t2);
			dyB_row[x] = dy * cell.coeff;
		}
	}
}

template <class image_type>
blurred_grid_22<image_type>::blurred_grid_22(image_type& src, float tau, blur_range range, int max_images)
	: blurred_grid<image_type>(src, tau, range, max_images, true) { }

template <class image_type>
void blurred_grid_22<image_type>::blur(const image1f& u, const image1f& v, image_type& B, image_type& duB, image_type& dvB, image_type& dxB, image_type& dyB)
{
	assert(dimensions_match(u, v));
	int width = u.width(), height = u.height();
	B.allocate(width, height); duB.allocate(width, height); dvB.allocate(width, height);
	dxB.allocate(width, height); dyB.allocate(width, height);

	float xu_max = width - 1.0f, yv_max = height - 1.0f;

	for (int y = 0; y < height; y++)
	{
		const float *u_row = u.row(y), *v_row = v.row(y);
		pixel_type *B_row = B.row(y), *duB_row = duB.row(y), *dvB_row = dvB.row(y);
		pixel_type *dxB_row = dxB.row(y), *dyB_row = dyB.row(y);
		blurred_image_cell cell;
		image_type *B11 = nullptr, *B21= nullptr, *B12 = nullptr, *B22 = nullptr;
		image_type *dxB11 = nullptr, *dxB21= nullptr, *dxB12 = nullptr, *dxB22 = nullptr;
		image_type *dyB11 = nullptr, *dyB21= nullptr, *dyB12 = nullptr, *dyB22 = nullptr;
		for (int x = 0; x < width; x++)
		{
			float u = u_row[x], v = v_row[x]; // w12(z)
			float xu = clamp(x + u, 0.0f, xu_max), yv = clamp(y + v, 0.0f, yv_max); // z + w12(z), used to warp image

			// Blur kernel is k[w12(z)tau1/2].
			u *= base::tau; v *= base::tau; // w12(z)tau1/2
			base::adjust_coord(u, v);

			if (cell.is_outside(u, v))
			{
				base::locate_cell(u, v, cell);
				B11 = base::images + cell.offset;
				B21 = base::images + cell.offset + 1;
				B12 = base::images + cell.offset + base::u_points.size();
				B22 = base::images + cell.offset + base::u_points.size() + 1;
				dxB11 = base::dx_images + cell.offset;
				dxB21 = base::dx_images + cell.offset + 1;
				dxB12 = base::dx_images + cell.offset + base::u_points.size();
				dxB22 = base::dx_images + cell.offset + base::u_points.size() + 1;
				dyB11 = base::dy_images + cell.offset;
				dyB21 = base::dy_images + cell.offset + 1;
				dyB12 = base::dy_images + cell.offset + base::u_points.size();
				dyB22 = base::dy_images + cell.offset + base::u_points.size() + 1;
			}

			u = clamp(u, cell.u1, cell.u2);
			v = clamp(v, cell.v1, cell.v2);

			float s1 = cell.u2 - u, s2 = u - cell.u1;
			float t1 = cell.v2 - v, t2 = v - cell.v1;

			pixel_type b11 = B11->interpolate(xu, yv), b21 = B21->interpolate(xu, yv);
			pixel_type b12 = B12->interpolate(xu, yv), b22 = B22->interpolate(xu, yv);

			pixel_type b = b11 * (s1 * t1);
			b += b21 * (s2 * t1);
			b += b12 * (s1 * t2);
			b += b22 * (s2 * t2);
			B_row[x] = b * cell.coeff;

			pixel_type du = (b21 - b11) * t1;
			du += (b22 - b12) * t2;
			duB_row[x] = du * (cell.coeff * base::tau);

			pixel_type dv = (b12 - b11) * s1;
			dv += (b22 - b21) * s2;
			dvB_row[x] = dv * (cell.coeff * base::tau);

			pixel_type dx11 = dxB11->interpolate(xu, yv), dx21 = dxB21->interpolate(xu, yv);
			pixel_type dx12 = dxB12->interpolate(xu, yv), dx22 = dxB22->interpolate(xu, yv);
			pixel_type dx = dx11 * (s1 * t1);
			dx += dx21 * (s2 * t1);
			dx += dx12 * (s1 * t2);
			dx += dx22 * (s2 * t2);
			dxB_row[x] = dx * cell.coeff;

			pixel_type dy11 = dyB11->interpolate(xu, yv), dy21 = dyB21->interpolate(xu, yv);
			pixel_type dy12 = dyB12->interpolate(xu, yv), dy22 = dyB22->interpolate(xu, yv);
			pixel_type dy = dy11 * (s1 * t1);
			dy += dy21 * (s2 * t1);
			dy += dy12 * (s1 * t2);
			dy += dy22 * (s2 * t2);
			dyB_row[x] = dy * cell.coeff;
		}
	}
}

#endif
