/*
 * filter.cpp
 *
 *  Created on: Oct 4, 2011
 *      Author: Travis Portz
 */

#include "filter.hpp"
#include "utility.hpp"
#include <cassert>

const float pi_f = 3.1415926535f;

const float gradient_kernel_data[5] = {1 / 12.0f, -8 / 12.0f, 0, 8 / 12.0f, -1 / 12.0f};
const kernel1d gradient_kernel(gradient_kernel_data, 5);

const float smoothing_kernel_data[9] = { 1 / 25.0f, 3 / 25.0f, 1 / 25.0f, 3 / 25.0f, 9 / 25.0f, 3 / 25.0f, 1 / 25.0f, 3 / 25.0f, 1 / 25.0f };
const kernel2d smoothing_kernel(smoothing_kernel_data, 3, 3);

void gaussian_kernel(float sigma, kernel1d& dst)
{
	int length = (fast_round(6.0f * sigma) + 1) | 1;
	int r = length / 2;

	float* data = new float[length];

	float scale = -1.0f / (2.0f * sigma * sigma);
	data[r] = 1.0f;
	for (int x = 1; x <= r; x++)
		data[r - x] = data[r + x] = std::exp(x * x * scale);

	float sum = 0.0f;
	for (int i = 0; i < length; i++)
		sum += data[i];
	scale = 1.0f / sum;
	for (int i = 0; i < length; i++)
		data[i] *= scale;

	dst.data = data;
	dst.length = length;
}

void motion_blur_kernel(float length, float theta, kernel2d& dst)
{
	if (length < 1.0f) length = 1.0f;
	if (theta < 0.0f) theta = pi_f + theta;
	const float line_width = 1.0f;
	const float half = (length - 1.0f) / 2;
	const float cos_theta = std::cos(theta), sin_theta = std::sin(theta);
	const float sign_x = (cos_theta > 0.0f) ? 1.0f : ((cos_theta < 0.0f) ? -1.0f : 0.0f);
	const float eps = length * std::numeric_limits<float>::epsilon();

	int sx = static_cast<int>(half * cos_theta + line_width * sign_x - eps);
	if (sx < 0) sx = -sx;
	int sy = static_cast<int>(half * sin_theta + line_width - eps);

	int width = 1 + 2 * sx;
	int height = 1 + 2 * sy;
	const int n = width * height;
	float* data = new float[n];
	std::fill(data, data + n, 0.0f);

	int x_start, x_end;
	if (cos_theta < 0.0f)
	{
		x_start = -sx;
		x_end = 0;
	}
	else
	{
		x_start = 0;
		x_end = sx;
	}

	for (int y = 0; y <= sy; y++)
	{
		float* row = data + (sy - y) * width;
		for (int x = x_start; x <= x_end; x++)
		{
			float dist_to_line = y * cos_theta - x * sin_theta; // Distance to line perpendicular to line.
			float radius = std::sqrt(static_cast<float>(x * x + y * y));
			if (radius >= half && std::abs(dist_to_line) <= line_width)
			{
				// Past end-point of line.
				float dist_to_end = std::abs(x * cos_theta + y * sin_theta) - half; // Distance to end-point parallel to line.
				dist_to_line = std::sqrt(dist_to_line * dist_to_line + dist_to_end * dist_to_end);
			}
			float value = line_width + std::numeric_limits<float>::epsilon() - std::abs(dist_to_line);
			if (value < 0.0f) value = 0.0f;
			row[sx + x] = value;
		}
	}

	if (cos_theta < 0.0f)
	{
		x_start = 0;
		x_end = sx;
	}
	else
	{
		x_start = -sx;
		x_end = 0;
	}

	for (int y = -sy; y <= 0; y++)
	{
		const float* src_row = data + (sy + y) * width;
		float* row = data + (sy - y) * width;
		for (int x = x_start; x <= x_end; x++)
		{
			row[sx + x] = src_row[sx - x];
		}
	}

	float sum = eps * length;
	for (int i = 0; i < n; i++)
		sum += data[i];
	float scale = 1.0f / sum;
	for (int i = 0; i < n; i++)
		data[i] *= scale;

	dst.data = data; dst.width = width; dst.height = height;
}

void motion_blur_kernel_oc(float u, float v, kernel2d& dst)
{
	assert(u != 0 || v != 0);
	u = std::abs(u); v = std::abs(v);
	const float length = std::sqrt(u * u + v * v);
	const float cos_theta = u / length, sin_theta = v / length;
	const float line_width = 1.0f;
	const float eps = length * std::numeric_limits<float>::epsilon();

	const float sign_x = (u > 0.0f) ? 1.0f : ((u < 0.0f) ? -1.0f : 0.0f);
	int sx = static_cast<int>(u + line_width * sign_x - eps);
	int x_offset = 0;
	if (sx < 0)
	{
		x_offset = sx;
		sx = -sx;
	}

	const float sign_y = (v > 0.0f) ? 1.0f : ((v < 0.0f) ? -1.0f : 0.0f);
	int sy = static_cast<int>(v + line_width * sign_y - eps);
	int y_offset = 0;
	if (sy < 0)
	{
		y_offset = sy;
		sy = -sy;
	}

	int width = 1 + sx;
	int height = 1 + sy;
	const int n = width * height;
	float* data = new float[n];
	std::fill(data, data + n, 0.0f);

	for (int yy = 0; yy <= sy; yy++)
	{
		int y = yy + y_offset;
		float* row = data + yy * width;
		for (int xx = 0; xx <= sx; xx++)
		{
			int x = xx + x_offset;
			float dist_to_line = y * cos_theta - x * sin_theta; // Distance to line perpendicular to line.
			float radius = std::sqrt(static_cast<float>(x * x + y * y));
			if (radius >= length && std::abs(dist_to_line) <= line_width)
			{
				// Past end-point of line.
				float dist_to_end = std::abs(x * cos_theta + y * sin_theta) - length; // Distance to end-point parallel to line.
				dist_to_line = std::sqrt(dist_to_line * dist_to_line + dist_to_end * dist_to_end);
			}
			float value = line_width + std::numeric_limits<float>::epsilon() - std::abs(dist_to_line);
			if (value < 0.0f) value = 0.0f;
			row[xx] = value;
		}
	}

	float sum = eps * length;
	for (int i = 0; i < n; i++)
		sum += data[i];
	float scale = 1.0f / sum;
	for (int i = 0; i < n; i++)
		data[i] *= scale;

	dst.data = data; dst.width = width; dst.height = height;
}
