/*
 * filter.hpp
 *
 *  Created on: Oct 4, 2011
 *      Author: Travis Portz
 */

#ifndef FILTER_HPP
#define FILTER_HPP

#include <limits>
#include "image.hpp"

struct kernel1d
{
	float* data;
	int length;

	kernel1d() : data(nullptr), length(0) { }
	kernel1d(const float* src, int length) : length(length)
	{
		data = new float[length];
		std::copy(src, src + length, data);
	}
	~kernel1d() { delete [] data; }
};

struct kernel2d
{
	float* data;
	int width, height;

	kernel2d() : data(nullptr), width(0), height(0) { }
	kernel2d(const float* src, int w, int h) : width(w), height(h)
	{
		int n = w * h;
		data = new float[n];
		std::copy(src, src + n, data);
	}
	~kernel2d() { delete [] data; }
};

extern const kernel1d gradient_kernel;
extern const kernel2d smoothing_kernel;
void gaussian_kernel(float sigma, kernel1d& dst);
void motion_blur_kernel(float length, float theta, kernel2d& dst);
// Creates off-center blur kernel going from (0,0) to (u,v). Must be used with filter_xy_oc.
void motion_blur_kernel_oc(float u, float v, kernel2d& dst);

// 2D filter
template <typename P>
void filter_xy(const image<P>& src, const kernel2d& kernel, image<P>& dst)
{
	dst.allocate(src);

	int x_max = static_cast<int>(src.width()) - 1;
	int y_max = static_cast<int>(src.height()) - 1;

	const float* kern_data = kernel.data;
	int kern_width = kernel.width, kern_height = kernel.height;
	int u_radius = kern_width / 2;
	int v_radius = kern_height / 2;

	for (int y = 0; y <= y_max; y++)
	{
		P* dst_row = dst.row(y);
		for (int x = 0; x <= x_max; x++)
		{
			P dst_pix(0);
			const float* kern = kern_data;
			for (int v = -v_radius; v <= v_radius; v++)
			{
				int yv = clamp(y + v, 0, y_max);
				for (int u = -u_radius; u <= u_radius; u++)
				{
					int xu = clamp(x + u, 0, x_max);
					float h =  *(kern++);
					dst_pix += src(xu, yv) * h;
				}
			}
			dst_row[x] = dst_pix;
		}
	}
}

// 1D horizontal filter
template <typename P>
void filter_x(const image<P>& src, const kernel1d& kernel, image<P>& dst)
{
	dst.allocate(src);

	int x_max = static_cast<int>(src.width()) - 1;
	int y_max = static_cast<int>(src.height()) - 1;

	const float* kern_data = kernel.data;
	int kern_width = kernel.length;
	int u_radius = kern_width / 2;

	for (int y = 0; y <= y_max; y++)
	{
		P* dst_row = dst.row(y);
		for (int x = 0; x <= x_max; x++)
		{
			P dst_pix(0);
			const float* kern = kern_data;
			for (int u = -u_radius; u <= u_radius; u++)
			{
				int xu = clamp(x + u, 0, x_max);
				float h =  *(kern++);
				dst_pix += src(xu, y) * h;
			}
			dst_row[x] = dst_pix;
		}
	}
}

// 1D vertical filter
template <typename P>
void filter_y(const image<P>& src, const kernel1d& kernel, image<P>& dst)
{
	dst.allocate(src);

	int x_max = static_cast<int>(src.width()) - 1;
	int y_max = static_cast<int>(src.height()) - 1;

	const float* kern_data = kernel.data;
	int kern_height = kernel.length;
	int v_radius = kern_height / 2;

	for (int y = 0; y <= y_max; y++)
	{
		P* dst_row = dst.row(y);
		for (int x = 0; x <= x_max; x++)
		{
			P dst_pix(0);
			const float* kern = kern_data;
			for (int v = -v_radius; v <= v_radius; v++)
			{
				int yv = clamp(y + v, 0, y_max);
				float h =  *(kern++);
				dst_pix += src(x, yv) * h;
			}
			dst_row[x] = dst_pix;
		}
	}
}

template <typename P>
void weighted_laplacian(const image<P>& src, const image<P>& weights, image<P>& dst)
{
	dst.allocate(src);

	int width = src.width(), height = src.height();
	int x_max = width - 1, y_max = height - 1;

	// Filter horizontally
	for (int y = 0; y < height; y++)
	{
		const P* src_row = src.row(y);
		const P* w_row = weights.row(y);
		P* dst_row = dst.row(y);
		P previous(0);
		for (int x = 0; x < x_max; x++)
		{
			P current = (src_row[x + 1] - src_row[x]) * w_row[x];
			dst_row[x] = previous - current;
			previous = current;
		}
		dst_row[x_max] = previous;
	}

	// Filter vertically
	static image<P> buffer;
	buffer.allocate(width, height);
	for (int y = 0; y < y_max; y++)
	{
		const P* src_row = src.row(y);
		const P* w_row = weights.row(y);
		P* buffer_row = buffer.row(y);
		for (int x = 0; x < width; x++)
		{
			buffer_row[x] = (src_row[x + width] - src_row[x]) * w_row[x];
		}
	}
	for (int y = 0; y < height; y++)
	{
		const P* buffer_row = buffer.row(y);
		P* dst_row = dst.row(y);
		for (int x = 0; x < width; x++)
		{
			if (y < y_max)
				dst_row[x] -= buffer_row[x];
			if (y > 0)
				dst_row[x] += buffer_row[x - width];
		}
	}
}

// Horizontal finite difference.
template <typename P>
void simple_dx(const image<P>& src, image<P>& dst)
{
	dst.allocate(src);

	int w = src.width(), h = src.height();
	int x_max = w - 1;
	for (int y = 0; y < h; y++)
	{
		const P* src_row = src.row(y);
		P* dst_row = dst.row(y);
		for (int x = 0; x < x_max; x++)
		{
			dst_row[x] = src_row[x + 1] - src_row[x];
		}
		dst_row[x_max] = dst_row[x_max - 1];
	}
}

// Vertical finite difference.
template <typename P>
void simple_dy(const image<P>& src, image<P>& dst)
{
	dst.allocate(src);

	int w = src.width(), h = src.height();
	int y_max = h - 1;
	for (int y = 0; y < y_max; y++)
	{
		const P* src_row = src.row(y);
		const P* src_row_next = src.row(y + 1);
		P* dst_row = dst.row(y);
		for (int x = 0; x < w; x++)
			dst_row[x] = src_row_next[x] - src_row[x];
	}
	P* last_row = dst.row(y_max);
	const P* prev_row = dst.row(y_max - 1);
	for (int x = 0; x < w; x++)
		last_row[x] = prev_row[x];
}

// Off-center kernel (origin is at corner based on signs of u and v)
template <typename P>
void filter_xy_oc(const image<P>& src, const kernel2d& kernel, float u, float v, image<P>& dst)
{
	dst.allocate(src);

	int x_max = static_cast<int>(src.width()) - 1;
	int y_max = static_cast<int>(src.height()) - 1;

	const float* kern_data = kernel.data;
	int kern_width = kernel.width, kern_height = kernel.height;

	int u_start = 0, u_stop = kern_width, du = 1;
	if (u > 0)
	{
		u_stop = -u_stop;
		du = -1;
	}
	int v_start = 0, v_stop = kern_height, dv = 1;
	if (v > 0)
	{
		v_stop = -v_stop;
		dv = -1;
	}

	for (int y = 0; y <= y_max; y++)
	{
		P* dst_row = dst.row(y);
		for (int x = 0; x <= x_max; x++)
		{
			P dst_pix(0);
			const float* kern = kern_data;
			for (int v = v_start; v != v_stop; v += dv)
			{
				int yv = clamp(y + v, 0, y_max);
				for (int u = u_start; u != u_stop; u += du)
				{
					int xu = clamp(x + u, 0, x_max);
					float h =  *(kern++);
					dst_pix += src(xu, yv) * h;
				}
			}
			dst_row[x] = dst_pix;
		}
	}
}

#endif /* FILTER_HPP */
