/*
 * flow.hpp
 *
 *  Created on: Aug 9, 2011
 *      Author: Travis Portz
 */

#ifndef FLOW_HPP
#define FLOW_HPP

#include <string>
#include <fstream>
#include <cassert>
#include "utility.hpp"
#include "image.hpp"

template <typename From, typename To>
inline To flow_convert_value(From from) { return static_cast<To>(from); }

template<>
inline int flow_convert_value<float, int>(float from) { return fast_round(from); }

template<>
inline short flow_convert_value<float, short>(float from) { return static_cast<short>(fast_round(from)); }

template<>
inline int flow_convert_value<double, int>(double from) { return fast_round(from); }

template <typename T>
struct flow
{
	T x, y;

	flow() : x(0), y(0) { }
	flow(T x, T y) : x(x), y(y) { }
	flow(const flow<T>& other) : x(other.x), y(other.y) { }

	template <typename U>
	explicit flow(const flow<U>& other) : x(flow_convert_value<U, T>(other.x)), y(flow_convert_value<U, T>(other.y)) { }

	flow<T>& operator=(const flow<T>& other)
	{
		x = other.x;
		y = other.y;
		return *this;
	}
	flow<T>& operator+=(const flow<T>& other)
	{
		x += other.x;
		y += other.y;
		return *this;
	}
	flow<T> operator+(const flow<T>& other) const
	{
		return flow<T>(x + other.x, y + other.y);
	}
	flow<T>& operator*=(float s)
	{
		x *= s;
		y *= s;
		return *this;
	}
	flow<T> operator*(float s) const
	{
		return flow<T>(x * s, y * s);
	}
};

template <typename T>
inline coord2& operator+=(coord2& c, const flow<T>& flow)
{
	c.x += flow_convert_value<T, int>(flow.x);
	c.y += flow_convert_value<T, int>(flow.y);
	return c;
}
template <typename T>
inline coord2f& operator+=(coord2f& c, const flow<T>& flow)
{
	c.x += flow_convert_value<T, float>(flow.x);
	c.y += flow_convert_value<T, float>(flow.y);
	return c;
}

template <typename T>
class flow_field
{
public:
	typedef flow<T> value_type;

	flow_field() : _width(0), _height(0), _data(nullptr), _owner(false)
	{ }

	flow_field(int width, int height) : _width(width), _height(height), _owner(true)
	{
		_data = new value_type[width * height];
	}
	// Shallow copy constructor
	flow_field(const flow_field<T>& other) : _width(other._width), _height(other._height), _data(other._data), _owner(false)
	{ }

	~flow_field()
	{
		if (_owner)
		{
			delete [] _data;
			_data = nullptr;
			_owner = false;
		}
	}

	void deep_copy(const flow_field<T>& other)
	{
		allocate(other._width, other._height);
		int n = _width * _height;
		for (int i = 0; i < n; i++)
		{
			_data[i] = other._data[i];
		}
	}

	void allocate(int width, int height)
	{
		if (width != _width || height != _height)
		{
			if (_owner)
				delete [] _data;
			_data = new value_type[width * height];
			_width = width;
			_height = height;
			_owner = true;
		}
	}

	inline int width() const { return _width; }
	inline int height() const { return _height; }

	inline value_type* data() { return _data; }
	inline const value_type* data() const { return _data; }

	inline value_type* row(int y) { return _data + _width * y; }
	inline const value_type* row(int y) const { return _data + _width * y; }

	inline value_type& at(int x, int y) { return _data[_width * y + x]; }
	inline const value_type& at(int x, int y) const { return _data[_width * y + x]; }
	inline value_type& at(coord2 c) { return _data[_width * c.y + c.x]; }
	inline const value_type& at(coord2 c) const { return _data[_width * c.y + c.x]; }

	inline const value_type at(coord2f c) const
	{
		float fx0 = std::floor(c.x);
		int x0 = static_cast<int>(fx0);
		int x1 = static_cast<int>(std::ceil(c.x));
		float fy0 = std::floor(c.y);
		int y0 = static_cast<int>(fy0);
		int y1 = static_cast<int>(std::ceil(c.y));
		float sx = c.x - fx0;
		value_type flow_y0 = at(x0, y0) * (1 - sx) + at(x1, y0) * sx;
		value_type flow_y1 = at(x0, y1) * (1 - sx) + at(x1, y1) * sx;
		float sy = c.y - fy0;
		return flow_y0 * (1 - sy) + flow_y1 * sy;
	}

	value_type interpolate(float x, float y) const
	{
		float fx0 = std::floor(x);
		int x0 = static_cast<int>(fx0);
		int x1 = static_cast<int>(std::ceil(x));
		float fy0 = std::floor(y);
		int y0 = static_cast<int>(fy0);
		int y1 = static_cast<int>(std::ceil(y));
		int offset = y0 * _width + x0;
		int dx = x1 - x0, dy = (y1 > y0) ? _width : 0;
		float sx = x - fx0;
		value_type flow_y0 = _data[offset] * (1.0f - sx) + _data[offset + dx] * sx;
		value_type flow_y1 = _data[offset + dy] * (1.0f - sx) + _data[offset + dy + dx] * sx;
		float sy = y - fy0;
		return flow_y0 * (1.0f - sy) + flow_y1 * sy;
	}

private:
	int _width, _height;
	value_type* _data;
	bool _owner;
};

template <typename From, typename To>
void flow_convert_field(const flow_field<From>& from, flow_field<To>& to)
{
	typedef flow<From> from_type;
	typedef flow<To> to_type;

	int width = from.width(), height = from.height();
	to.allocate(width, height);

	const from_type* src = from.data();
	to_type* dst = to.data();
	int n = width * height;
	for (int i = 0; i < n; i++)
	{
		dst[i] = static_cast<to_type>(src[i]);
	}
}

template <typename T>
void flow_read_field(const std::string& filename, flow_field<T>& field)
{
	std::ifstream fin(filename.c_str(), std::ios::binary);
	fin.seekg(8);
	int width, height;
	fin.read(reinterpret_cast<char*>(&width), sizeof(int));
	fin.read(reinterpret_cast<char*>(&height), sizeof(int));
	field.allocate(width, height);
	fin.read(reinterpret_cast<char*>(field.data()), sizeof(flow<T>) * width * height);
	fin.close();
}
template <typename T>
void flow_write_field(const std::string& filename, flow_field<T>& field)
{
	std::ofstream fout(filename.c_str(), std::ios::binary);
	fout.write("OPTIFLOW", 8);
	int width = field.width(), height = field.height();
	fout.write(reinterpret_cast<const char*>(&width), sizeof(int));
	fout.write(reinterpret_cast<const char*>(&height), sizeof(int));
	fout.write(reinterpret_cast<const char*>(field.data()), sizeof(flow<T>) * width * height);
	fout.close();
}

template <typename T>
void flow_concatenate_fields(const flow_field<T>& field1, const flow_field<T>& field2, flow_field<T>& result)
{
	int width = field1.width(), height = field1.height();
	assert(field2.width() == width && field2.height() == height);
	result.allocate(width, height);

	for (int y = 0; y < height; y++)
	{
		const flow<T>* src_ptr = field1.row(y);
		flow<T>* dst_ptr = result.row(y);
		for (int x = 0; x < width; x++)
		{
			coord2 c(x, y);
			flow<T> flow1 = *(src_ptr++);
			c += flow1;
			if (c.in_bounds(width, height))
			{
				flow1 += field2.at(c);
			}
			*(dst_ptr++) = flow1;
		}
	}
}

template <typename T>
void flow_to_matlab(const flow_field<T>& field, mxArray*& uv)
{
	mwSize dims[3] = { field.height(), field.width(), 2 };
	uv = mxCreateNumericArray(3, dims, mxDOUBLE_CLASS, mxREAL);
	double* uv_ptr = static_cast<double*>(mxGetData(uv));
	// Horizontal component.
	for (int x = 0; x < dims[1]; x++)
		for (int y = 0; y < dims[0]; y++)
			*(uv_ptr++) = static_cast<double>(field.at(x, y).x);
	// Vertical component.
	for (int x = 0; x < dims[1]; x++)
		for (int y = 0; y < dims[0]; y++)
			*(uv_ptr++) = static_cast<double>(field.at(x, y).y);
}

template <typename T>
void flow_from_matlab(const mxArray* uv, flow_field<T>& field)
{
	const mwSize* dims = mxGetDimensions(uv);
	// Input should be de-interleaved UV, double format.
	assert(dims[2] == 2 && mxGetClassID(uv) == mxDOUBLE_CLASS);
	field.allocate(dims[1], dims[0]); // Column-major to row-major.
	const double* uv_ptr = static_cast<const double*>(mxGetData(uv));
	// Horizontal component.
	for (int x = 0; x < dims[1]; x++)
		for (int y = 0; y < dims[0]; y++)
			field.at(x, y).x = static_cast<float>(*(uv_ptr++));
	// Vertical component.
	for (int x = 0; x < dims[1]; x++)
		for (int y = 0; y < dims[0]; y++)
			field.at(x, y).y = static_cast<float>(*(uv_ptr++));
}

template <typename T>
void flow_to_images(const flow_field<T>& field, image<T>& u, image<T>& v)
{
	int width = field.width(), height = field.height();
	u.allocate(width, height);
	v.allocate(width, height);

	for (int y = 0; y < height; y++)
	{
		const flow<T>* src_ptr = field.row(y);
		T* u_ptr = u.row(y);
		T* v_ptr = v.row(y);
		for (int x = 0; x < width; x++)
		{
			u_ptr[x] = src_ptr[x].x;
			v_ptr[x] = src_ptr[x].y;
		}
	}
}

template <typename T>
void flow_from_images(const image<T>& u, const image<T>& v, flow_field<T>& field)
{
	int width = u.width(), height = u.height();
	field.allocate(width, height);

	for (int y = 0; y < height; y++)
	{
		flow<T>* dst_ptr = field.row(y);
		const T* u_ptr = u.row(y);
		const T* v_ptr = v.row(y);
		for (int x = 0; x < width; x++)
		{
			dst_ptr[x] = flow<T>(u_ptr[x], v_ptr[x]);
		}
	}
}

template <typename T>
void flow_resize(const flow_field<T>& src, flow_field<T>& dst, int dst_w, int dst_h)
{
	int src_w = src.width();
	int src_h = src.height();

	double x_scale = dst_w / static_cast<double>(src_w);
	double y_scale = dst_h / static_cast<double>(src_h);
	double inv_x_scale = 1.0 / x_scale;
	double inv_y_scale = 1.0 / y_scale;

	dst.allocate(dst_w, dst_h); // Allocate destination image with proper size

	double x_max = src_w - 1.0;
	double y_max = src_h - 1.0;

	// Resampling coordinate offsets to center interpolation kernel
	double x_offset = 0.5 * ((src_w - 1) - (dst_w - 1) * inv_x_scale);
	double y_offset = 0.5 * ((src_h - 1) - (dst_h - 1) * inv_y_scale);

	for (int j = 0; j < dst_h; j++)
	{
		double y = clamp(y_offset + j * inv_y_scale, 0.0, y_max);
		flow<T>* dst_it = dst.row(j);
		for (int i = 0; i < dst_w; i++)
		{
			double x = clamp(x_offset + i * inv_x_scale, 0.0, x_max);
			flow<T> f = src.interpolate(x, y);
			f.x *= x_scale; f.y *= y_scale;
			dst_it[i] = f;
		}
	}
}

struct flow_index
{
	int from, to;
	flow_index(int from, int to) : from(from), to(to) { }
	bool operator==(const flow_index& other) const { return (from == other.from && to == other.to); }
	bool operator<(const flow_index& other) const
	{
		if (from != other.from)
			return from < other.from;
		else
			return to < other.to;
	}
};


typedef flow<short> flow16;
typedef flow_field<short> flow16_field;
typedef flow<float> flow32f;
typedef flow_field<float> flow32f_field;

#endif /* FLOW_HPP */
