/*
 * image.hpp
 *
 *  Created on: Aug 5, 2011
 *      Author: Travis Portz
 */

#ifndef IMAGE_HPP
#define IMAGE_HPP

#include "mex.h"
#include <cmath>
#include <vector>
#include <algorithm>
#include <cassert>
#include "utility.hpp"
#include "pixel.hpp"
#include "locator.hpp"

template <typename P>
class image
{
public:
	typedef P pixel_type;

	image() : _data(nullptr), _width(0), _height(0), _owner(false) { }
	image(int width, int height) : _width(width), _height(height), _owner(true)
	{
		_data = new P[width * height];
	}
	image(const P* data, int width, int height) : _width(width), _height(height), _owner(true)
	{
		int n = width * height;
		_data = new P[n];
		std::copy(data, data + n, _data);
	}
	image(const image<P>& other) : _width(width), _height(height), _owner(true)
	{
		int n = width * height;
		_data = new P[n];
		std::copy(other._data, other._data + n, _data);
	}
	~image() { if (_owner) delete [] _data; }

	void allocate(int width, int height)
	{
		if (!_owner || (_width != width && _height != height))
		{
			if (_owner)
				delete [] _data;
			else
				_owner = true;
			_data = new P[width * height];
			_width = width;
			_height = height;
		}
	}

	template <typename P2>
	inline void allocate(const image<P2>& other)
	{
		allocate(other.width(), other.height());
	}

	// Initializes image to all zeros
	void zero()
	{
		P zero(0);
		std::fill(_data, _data + _width * _height, zero);
	}

	void deep_copy(const image<P>& other)
	{
		int w = other.width(), h = other.height();
		allocate(w, h);
		std::copy(other._data, other._data + w * h, _data);
	}

	// Points this image at the data of the other image.
	void shallow_copy(image<P>& other)
	{
		if (_owner)
		{
			delete [] _data;
			_owner = false;
		}
		_data = other._data;
		_width = other._width;
		_height = other._height;
	}

	int width() const { return _width; }
	int height() const { return _height; }
	P* data() { return _data; }
	const P* data() const { return _data; }

	P* row(int y) { return _data + y * _width; }
	const P* row(int y) const { return _data + y * _width; }

	locator<P> at(int x, int y)
	{
		return locator<P>(_data + y * _width + x, _width);
	}

	P operator()(int x, int y) const
	{
		return _data[y * _width + x];
	}
	P& operator()(int x, int y)
	{
		return _data[y * _width + x];
	}

	// Bilinear interpolation. Does not do extrapolation.
	P interpolate(float x, float y) const
	{
		float fx0 = std::floor(x);
		int x0 = static_cast<int>(fx0);
		int x1 = static_cast<int>(std::ceil(x));
		float fy0 = std::floor(y);
		int y0 = static_cast<int>(fy0);
		int y1 = static_cast<int>(std::ceil(y));
		int offset = y0 * _width + x0;
		int dx = x1 - x0, dy = (y1 > y0) ? _width : 0;
		float sx = x - fx0;
		P pix_y0 = _data[offset] * (1.0f - sx) + _data[offset + dx] * sx;
		P pix_y1 = _data[offset + dy] * (1.0f - sx) + _data[offset + dy + dx] * sx;
		float sy = y - fy0;
		return pix_y0 * (1.0f - sy) + pix_y1 * sy;
	}

	void swap_with(image<P>& other)
	{
		std::swap(_data, other._data);
		std::swap(_width, other._width);
		std::swap(_height, other._height);
		std::swap(_owner, other._owner);
	}

	template <typename Scalar>
	void multiply_with(Scalar a)
	{
		int n = _width * _height;
		P* data = _data;
		for (int i = 0; i < n; i++)
			data[i] *= a;
	}

	void average_with(const image<P>& other)
	{
		int n = _width * _height;
		P* data = _data;
		const P* other_data = other._data;
		for (int i = 0; i < n; i++)
		{
			data[i] = (data[i] + other_data[i]) * 0.5;
		}
	}

	void add(P p)
	{
		int n = _width * _height;
		P* data = _data;
		for (int i = 0; i < n; i++)
			data[i] += p;
	}

	void add(const image<P>& other)
	{
		int n = _width * _height;
		P* data = _data;
		const P* other_data = other._data;
		for (int i = 0; i < n; i++)
			data[i] += other_data[i];
	}

	template <typename Scalar>
	void add(const image<P>& other, Scalar ratio)
	{
		int n = _width * _height;
		P* data = _data;
		const P* other_data = other._data;
		for (int i = 0; i < n; i++)
			data[i] += other_data[i] * ratio;
	}

	void subtract(const image<P>& other)
	{
		int n = _width * _height;
		P* data = _data;
		const P* other_data = other._data;
		for (int i = 0; i < n; i++)
			data[i] -= other_data[i];
	}

private:
	P* _data;
	int _width, _height;
	bool _owner;
};

typedef image<pixel3c> image3c;
typedef image<pixel3f> image3f;
typedef image<float> image1f;

// Load image from Matlab array.
void load_image(const mxArray* data, image3f& img);
// Save image to Matlab array.
void save_image(mxArray* data, const image3f& img);

template <typename P>
void resize(const image<P>& src, image<P>& dst, double scale)
{
	double inv_scale = 1.0 / scale;

	int src_w = src.width();
	int src_h = src.height();
	int dst_w = static_cast<int>(src_w * scale);
	int dst_h = static_cast<int>(src_h * scale);
	dst.allocate(dst_w, dst_h); // Allocate destination image with proper size

	double x_max = src_w - 1.0;
	double y_max = src_h - 1.0;

	// Resampling coordinate offsets to center interpolation kernel
	double x_offset = 0.5 * ((src_w - 1) - (dst_w - 1) * inv_scale);
	double y_offset = 0.5 * ((src_h - 1) - (dst_h - 1) * inv_scale);

	for (int j = 0; j < dst_h; j++)
	{
		double y = clamp(y_offset + j * inv_scale, 0.0, y_max);
		P* dst_it = dst.row(j);
		for (int i = 0; i < dst_w; i++)
		{
			double x = clamp(x_offset + i * inv_scale, 0.0, x_max);
			dst_it[i] = src.interpolate(x, y);
		}
	}
}

template <typename P>
void resize(const image<P>& src, image<P>& dst, int dst_w, int dst_h)
{
	int src_w = src.width();
	int src_h = src.height();

	double inv_x_scale = src_w / static_cast<double>(dst_w);
	double inv_y_scale = src_h / static_cast<double>(dst_h);

	dst.allocate(dst_w, dst_h); // Allocate destination image with proper size

	double x_max = src_w - 1.0;
	double y_max = src_h - 1.0;

	// Resampling coordinate offsets to center interpolation kernel
	double x_offset = 0.5 * ((src_w - 1) - (dst_w - 1) * inv_x_scale);
	double y_offset = 0.5 * ((src_h - 1) - (dst_h - 1) * inv_y_scale);

	for (int j = 0; j < dst_h; j++)
	{
		double y = clamp(y_offset + j * inv_y_scale, 0.0, y_max);
		P* dst_it = dst.row(j);
		for (int i = 0; i < dst_w; i++)
		{
			double x = clamp(x_offset + i * inv_x_scale, 0.0, x_max);
			dst_it[i] = src.interpolate(x, y);
		}
	}
}

// Build pyramid using pyramid[0] as finest level
template <typename P>
void build_pyramid(image<P>* pyramid, int levels, double scale_factor)
{
	for (int i = 1; i < levels; i++)
	{
		resize(pyramid[i - 1], pyramid[i], scale_factor);
	}
}

template <typename P1, typename P2>
inline bool dimensions_match(const image<P1>& a, const image<P2>& b)
{
	return (a.width() == b.width() && a.height() == b.height());
}

template <typename T>
void desaturate(const image<pixel<T, 3> >& src, image<T>& dst)
{
	dst.allocate(src);
	int n = src.width() * src.height();
	const pixel<T, 3>* src_data = src.data();
	T* dst_data = dst.data();
	for (int i = 0; i < n; i++)
	{
		pixel<T, 3> p = src_data[i];
		dst_data[i] = static_cast<T>(0.2989 * p[0] + 0.5870 * p[1] + 0.1140 * p[2]);
	}
}

// Convert to single channel image by summing channels.
template <typename T, unsigned C>
void collapse(const image<pixel<T, C> >& src, image<T>& dst)
{
	dst.allocate(src);
	int n = src.width() * src.height();
	const pixel<T, C>* src_data = src.data();
	T* dst_data = dst.data();
	for (int i = 0; i < n; i++)
	{
		dst_data[i] = src_data[i].collapse();
	}
}

template <> void desaturate<float>(const image3f& src, image1f& dst);

#endif /* IMAGE_HPP */
